"""Main module."""
# import dependencies
from ast import expr
from dataclasses import InitVar
from pyomo.environ import ConcreteModel, Set, Var, Objective, minimize, maximize, Constraint ,NonNegativeReals,PositiveReals
import numpy as np
import pandas as pd
from .constant import CET_ADDI, ORIENT_IO, ORIENT_OO, RTS_VRS, RTS_CRS, OPT_DEFAULT, OPT_LOCAL
from .utils import tools


class DEA:
    """Data Envelopment Analysis (DEA)
    """

    def __init__(self, y, x, orient, rts, yref=None, xref=None):
        """DEA: Envelopment problem

        Args:
            y (float): output variable.
            x (float): input variables.
            orient (String): ORIENT_IO (input orientation) or ORIENT_OO (output orientation)
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        # Initialize DEA model
        self.orient = orient
        self.rts = rts
        self.y, self.x = tools.assert_valid_mupltiple_y_data(y, x)
        self.orient = orient
        self.rts = rts

        if type(yref) != type(None):
            self.yref, self.xref = tools.assert_valid_reference_data(self.y, self.x, yref, xref)
        else:
            self.yref, self.xref = self.y, self.x

  ## 当前被评价决策单元的序号 self.x[I0]
        # Initialize sets
        # print(self.xref)
        self.I = range(len(self.x))          ## I 是 被评价决策单元的数量
        self.__modeldict = {}
        for i in self.I:
            self.I0 = i

            self.__model__ = ConcreteModel()

            self.__model__.I2 = Set(initialize=range(len(self.xref)))      ## I2 是 参考决策单元的数量
            self.__model__.K = Set(initialize=range(len(self.x[0])))          ## K 是投入个数
            self.__model__.L = Set(initialize=range(len(self.y[0])))           ## L 是产出个数 被评价单元和参考单元的K，L一样

            # Initialize variable
            self.__model__.theta = Var(Set(initialize=range(1)),bounds=(None, None), doc='efficiency')
            self.__model__.lamda = Var(self.__model__.I2, bounds=(0.0, None), doc='intensity variables')

            # Setup the objective function and constraints
            if self.orient == ORIENT_IO:
                self.__model__.objective = Objective(
                    rule=self.__objective_rule(), sense=minimize, doc='objective function')
            else:
                self.__model__.objective = Objective(
                    rule=self.__objective_rule(), sense=maximize, doc='objective function')

            self.__model__.input = Constraint(
                self.__model__.K, rule=self.__input_rule(), doc='input constraint')
            self.__model__.output = Constraint(
                self.__model__.L, rule=self.__output_rule(), doc='output constraint')

            if self.rts == RTS_VRS:
                self.__model__.vrs = Constraint(
                     rule=self.__vrs_rule(), doc='various return to scale rule')

            self.__modeldict[i] = self.__model__

        # # Optimize model


    def __objective_rule(self):
        """Return the proper objective function"""

        def objective_rule(model):
            return model.theta[0]*1  + sum(model.lamda[i2] *0 for i2 in model.I2)

        return objective_rule


    def __input_rule(self):
        """Return the proper input constraint"""
        if self.orient == ORIENT_OO:
            def input_rule(model, k):
                return sum(model.lamda[i2] * self.xref[i2,k] for i2 in model.I2) <= self.x[self.I0,k]
        elif self.orient == ORIENT_IO:
            def input_rule(model, k):
                return sum(model.lamda[i2] * self.xref[i2,k] for i2 in model.I2) <= model.theta * self.x[self.I0,k]
        return input_rule



    def __output_rule(self):
        """Return the proper output constraint"""
        if self.orient == ORIENT_OO:
            def output_rule(model, l):
                return sum(model.lamda[i2] * self.yref[i2][l] for i2 in model.I2) >=model.theta * self.y[self.I0][l]
        elif self.orient == ORIENT_IO:
            def output_rule(model, l):
                return sum(model.lamda[i2] * self.yref[i2][l] for i2 in model.I2) >= self.y[self.I0][l]
        return output_rule

    def __vrs_rule(self):
        def vrs_rule(model):
            return sum(model.lamda[ i2] for i2 in model.I2) == 1

        return vrs_rule

    def optimize(self,  solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        
        problem_status,optimization_status,theta,lamda = {},{},{},{}
        for ind, problem in list(self.__modeldict.items()):
            problem_status[ind], optimization_status[ind] = tools.optimize_model(problem, CET_ADDI, solver)
            theta[ind],lamda[ind]=np.asarray(list(problem.theta[:].value)),np.asarray(list(problem.lamda[:].value))
        
        theta = pd.DataFrame(theta).T
        theta.columns = ["theta"]
        theta["eff"]=1/theta["theta"]
        lamda =pd.DataFrame(lamda) 
        lamda.columns = lamda.columns.map(lambda x : "Dum"+ str(x)+"'s lambda" ) 
        return problem_status,optimization_status,theta,lamda

    def info(self, dmu = "all"):
        """Show the infomation of the lp model

        Args:
            dmu (string): The solver chosen for optimization. Default is "all".
        """
        if dmu =="all":
            for ind, problem in list(self.__modeldict.items()):
                print(ind,"\n",problem.pprint())
 
        print(self.__modeldict[int(dmu)].pprint())




class RM(DEA):
    def __init__(self,  y, x,   orient, rts, yref=None, xref=None):
        """DEA: Directional distance function

        Args:
            y (float): output variable. 
            x (float): input variables.
            orient (String): ORIENT_IO (input orientation) or ORIENT_OO (output orientation)
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
        """
        # Initialize RM model

        self.orient = orient
        self.rts = rts
        self.y, self.x = tools.assert_valid_mupltiple_y_data(y, x)
        self.orient = orient
        self.rts = rts

        if type(yref) != type(None):
            self.yref, self.xref = tools.assert_valid_reference_data(self.y, self.x, yref, xref)
        else:
            self.yref, self.xref = self.y, self.x

  ## 当前被评价决策单元的序号 self.x[I0]
        # Initialize sets
        # print(self.xref)
        self.I = range(len(self.x))          ## I 是 被评价决策单元的数量
        self.__modeldict = {}
        for i in self.I:
            self.I0 = i

            self.__model__ = ConcreteModel()

            self.__model__.I2 = Set(initialize=range(len(self.xref)))      ## I2 是 参考决策单元的数量
            self.__model__.K = Set(initialize=range(len(self.x[0])))          ## K 是投入个数
            self.__model__.L = Set(initialize=range(len(self.y[0])))           ## L 是产出个数 被评价单元和参考单元的K，L一样

            # Initialize variable
            if self.orient == ORIENT_OO:
                self.__model__.theta = Var(self.__model__.L,bounds=(None, None), doc='efficiency')
            else:
                self.__model__.theta = Var(self.__model__.K,bounds=(None, None), doc='efficiency')
            
            self.__model__.lamda = Var(self.__model__.I2, bounds=(0.0, None), doc='intensity variables')

            # Setup the objective function  
            if self.orient == ORIENT_OO:
                self.__model__.objective = Objective(
                    rule=self.__objective_rule(), sense=maximize, doc='objective function')
            else:
                self.__model__.objective = Objective(
                    rule=self.__objective_rule(), sense=minimize, doc='objective function')
           
            # Setup the Constraint function  
            self.__model__.input = Constraint(
                self.__model__.K, rule=self.__input_rule(), doc='input constraint')
            self.__model__.output = Constraint(
                self.__model__.L, rule=self.__output_rule(), doc='output constraint')

            if self.rts == RTS_VRS:
                self.__model__.vrs = Constraint(rule=self.__vrs_rule(), doc='various return to scale rule')

            self.__modeldict[i] = self.__model__

        # # Optimize model



    def __objective_rule(self):
        """Return the proper objective function"""
        if self.orient == ORIENT_OO:
            def objective_rule(model):
                return sum(model.theta[l]*1 for l in model.L) / len(model.L)
        else:
            def objective_rule(model):
                return sum(model.theta[k]*1 for k in model.K) / len(model.K)
        return objective_rule


    def __input_rule(self):
        """Return the proper input constraint"""
        if self.orient == ORIENT_OO:
            def input_rule(model, k):
                return sum(model.lamda[i2] * self.xref[i2,k] for i2 in model.I2) <= self.x[self.I0,k]
        elif self.orient == ORIENT_IO:
            def input_rule(model, k):
                return sum(model.lamda[i2] * self.xref[i2,k] for i2 in model.I2) <= model.theta[k] * self.x[self.I0,k]
        return input_rule

    def __output_rule(self):
        """Return the proper output constraint"""
        if self.orient == ORIENT_OO:
            def output_rule(model, l):
                return sum(model.lamda[i2] * self.yref[i2][l] for i2 in model.I2) >=model.theta[l] * self.y[self.I0][l]
        elif self.orient == ORIENT_IO:
            def output_rule(model, l):
                return sum(model.lamda[i2] * self.yref[i2][l] for i2 in model.I2) >= self.y[self.I0][l]
        return output_rule


    def __vrs_rule(self):
        def vrs_rule(model):
            return sum(model.lamda[ i2] for i2 in model.I2) == 1

        return vrs_rule


    def optimize(self,  solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        
        problem_status,optimization_status,theta,lamda = {},{},{},{}
        for ind, problem in list(self.__modeldict.items()):
            problem_status[ind], optimization_status[ind] = tools.optimize_model(problem, CET_ADDI, solver)
            theta[ind],lamda[ind]=np.asarray(list(problem.theta[:].value)),np.asarray(list(problem.lamda[:].value))
        
        theta = pd.DataFrame(theta).T
        theta.columns = theta.columns.map(lambda x : "Input"+ str(x)+"'s theta" ) 
        for index,col in enumerate( theta.columns):
            theta["Input{}'s eff".format(index)]=theta[col]

        lamda =pd.DataFrame(lamda) 
        lamda.columns = lamda.columns.map(lambda x : "Dum"+ str(x)+"'s lambda" ) 
        return problem_status,optimization_status,theta,lamda

    def info(self, dmu = "all"):
        """Show the infomation of the lp model

        Args:
            dmu (string): The solver chosen for optimization. Default is "all".
        """
        if dmu =="all":
            for ind, problem in list(self.__modeldict.items()):
                print(ind,"\n",problem.pprint())
 
        print(self.__modeldict[int(dmu)].pprint())





class DDF(DEA):
    def __init__(self,  y, x, b=None, gy=[1], gx=[1], gb=None, rts=RTS_VRS, yref=None, xref=None, bref=None):
        """DEA: Directional distance function

        Args:
            y (float): output variable. 
            x (float): input variables.
            b (float), optional): undesirable output variables. Defaults to None.
            gy (list, optional): output directional vector. Defaults to [1].
            gx (list, optional): input directional vector. Defaults to [1].
            gb (list, optional): undesirable output directional vector. Defaults to None.
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
            bref (String, optional): reference undesirable output. Defaults to None.
        """
        # Initialize DEA model


        self.y, self.x, self.b, self.gy, self.gx, self.gb = tools.assert_valid_direciontal_data(y, x, b, gy, gx, gb)
        self.rts = rts

        if type(yref) != type(None):
            self.yref, self.xref, self.bref = tools.assert_valid_reference_data_with_bad_outputs(
                self.y, self.x, self.b, yref, xref, bref)
        else:
            self.yref, self.xref, self.bref = self.y, self.x, self.b

        self.I = range(len(self.x))          ## I 是 被评价决策单元的数量
        self.__modeldict = {}
        for i in self.I:
            self.I0 = i                                                 ## I 是 被评价决策单元的数量 

            self.__model__ = ConcreteModel()

            # Initialize sets
            self.__model__.I2 = Set(initialize=range(len(self.xref)))      ## I2 是 参考决策单元的数量
            self.__model__.K = Set(initialize=range(len(self.x[0])))          ## K 是投入个数
            self.__model__.L = Set(initialize=range(len(self.y[0])))           ## L 是产出个数 被评价单元和参考单元的K，L一样
            if type(b) != type(None):
                self.__model__.B = Set(initialize=range(len(self.b[0])))   ## B 是 非期望产出个数  
    
            # Initialize variable
            self.__model__.theta = Var(Set(initialize=range(1)),bounds=(0.0, None), doc='directional distance')

            self.__model__.lamda = Var(self.__model__.I2, bounds=(0.0, None), doc='intensity variables')

            # Setup the objective function and constraints
            self.__model__.objective = Objective(rule=self._DEA__objective_rule(), sense=maximize, doc='objective function')
            self.__model__.input = Constraint(self.__model__.K,  rule=self.__input_rule(), doc='input constraint')
            self.__model__.output = Constraint(self.__model__.L,  rule=self.__output_rule(), doc='output constraint')

            if type(b) != type(None):
                self.__model__.undesirable_output = Constraint(self.__model__.B, rule=self.__undesirable_output_rule(), doc='undesirable output constraint')

            if self.rts == RTS_VRS:
                self.__model__.vrs = Constraint(rule=self.__vrs_rule(), doc='various return to scale rule')
            
            self.__modeldict[i] = self.__model__

        # Optimize model

    def __input_rule(self):
        """Return the proper input constraint"""
        def input_rule(model, k):
            return sum(model.lamda[i2] * self.xref[i2][k] for i2 in model.I2) <= (1+model.theta[0]*self.gx[k])*self.x[self.I0][k]
        return input_rule

    def __output_rule(self):
        """Return the proper output constraint"""
        def output_rule(model, l):
            return sum(model.lamda[i2] * self.yref[i2,l] for i2 in model.I2) >= (1+model.theta[0]*self.gy[l])*self.y[self.I0][l]
        return output_rule

    def __undesirable_output_rule(self):
        """Return the proper undesirable output constraint"""
        def undesirable_output_rule(model, b):
            return sum(model.lamda[i2] * self.bref[i2,b] for i2 in model.I2) == (1+model.theta[0]*self.gb[b])*self.b[self.I0][b]
        return undesirable_output_rule



    def optimize(self,  solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        
        problem_status,optimization_status,theta,lamda = {},{},{},{}
        for ind, problem in list(self.__modeldict.items()):
            problem_status[ind], optimization_status[ind] = tools.optimize_model(problem, CET_ADDI, solver)
            theta[ind],lamda[ind]=np.asarray(list(problem.theta[:].value)),np.asarray(list(problem.lamda[:].value))
        
        theta = pd.DataFrame(theta).T
        theta.columns = ["Dirdectional distance function"] 


        lamda =pd.DataFrame(lamda) 
        lamda.columns = lamda.columns.map(lambda x : "Dum"+ str(x)+"'s lambda" ) 
        return problem_status,optimization_status,theta,lamda

    def info(self, dmu = "all"):
        """Show the infomation of the lp model

        Args:
            dmu (string): The solver chosen for optimization. Default is "all".
        """
        if dmu =="all":
            for ind, problem in list(self.__modeldict.items()):
                print(ind,"\n",problem.pprint())
 
        print(self.__modeldict[int(dmu)].pprint())




class NDDF(DEA):
    def __init__(self,  y, x, b=None, gy=[1], gx=[1], gb=None, weight =None,rts=RTS_VRS, yref=None, xref=None, bref=None):
        """DEA: Directional distance function

        Args:
            y (float): output variable. 
            x (float): input variables.
            b (float), optional): undesirable output variables. Defaults to None.
            gy (list, optional): output directional vector. Defaults to [1].
            gx (list, optional): input directional vector. Defaults to [1].
            gb (list, optional): undesirable output directional vector. Defaults to None.
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
            bref (String, optional): reference undesirable output. Defaults to None.
        """
        # Initialize DEA model
        self.b = b
        self.y, self.x, self.b, self.gy, self.gx, self.gb = tools.assert_valid_direciontal_data(y, x, b, gy, gx, gb)
        self.rts = rts
        if type(yref) != type(None):
            self.yref, self.xref, self.bref = tools.assert_valid_reference_data_with_bad_outputs(self.y, self.x, self.b, yref, xref, bref)
        else:
            self.yref, self.xref, self.bref = self.y, self.x, self.b
        
        if type(weight) != type(None):
            self.weight= tools.assert_valid_weight(self.y, self.x, self.b,weight)
        else:
            self.weight=[]
            if type(b) != type(None):
                for _ in range(len(self.x[0])):
                    self.weight.append(1/3/len(self.x[0]))
                for _ in range(len(self.y[0])):
                    self.weight.append(1/3/len(self.y[0]))
                for _ in range(len(self.b[0])):
                    self.weight.append(1/3/len(self.b[0]))  
            else:
                for _ in range(len(self.x[0])):
                    self.weight.append(1/2/len(self.x[0]))
                for _ in range(len(self.y[0])):
                    self.weight.append(1/2/len(self.y[0]))

        self.iweight = self.weight[0:len(self.x[0])]
        self.oweight = self.weight[len(self.x[0]):len(self.x[0])+len(self.y[0])]
        if type(b) != type(None):
            self.bweight = self.weight[len(self.x[0])+len(self.y[0]):len(self.x[0])+len(self.y[0])+len(self.b[0])]
        self.I = range(len(self.x))          ## I 是 被评价决策单元的数量
        self.__modeldict = {}
        for i in self.I:
            # print(self.gx,self.gy,self.gb,self.iweight,self.oweight,self.weight)
            self.I0 = i                                                 ## I 是 被评价决策单元的数量 

            self.__model__ = ConcreteModel()

            # Initialize sets
            self.__model__.I2 = Set(initialize=range(len(self.xref)))      ## I2 是 参考决策单元的数量
            self.__model__.K = Set(initialize=range(len(self.x[0])))          ## K 是投入个数
            self.__model__.L = Set(initialize=range(len(self.y[0])))           ## L 是产出个数 被评价单元和参考单元的K，L一样
            if type(b) != type(None):
                self.__model__.B = Set(initialize=range(len(self.b[0])))   ## B 是 非期望产出个数  


            # Initialize variable
            
            self.__model__.thetax = Var(self.__model__.K,initialize=1,bounds=(0.0, None), doc='scale factor x')
            self.__model__.thetay = Var(self.__model__.L,initialize=1,bounds=(0.0, None), doc='scale factor y')
            if type(b) != type(None):
                self.__model__.thetab = Var(self.__model__.B,initialize=1,bounds=(0.0, None), doc='scale factor B')
            self.__model__.lamda = Var(self.__model__.I2, bounds=(0.0, None), doc='intensity variables')

            # Setup the objective function and constraints
            self.__model__.objective = Objective(rule=self.__NDDF__objective_rule(), sense=maximize, doc='objective function')
            self.__model__.input = Constraint(self.__model__.K,  rule=self.__input_rule(), doc='input constraint')
            self.__model__.output = Constraint(self.__model__.L,  rule=self.__output_rule(), doc='output constraint')

            if type(b) != type(None):
                self.__model__.undesirable_output = Constraint(self.__model__.B, rule=self.__undesirable_output_rule(), doc='undesirable output constraint')

            if self.rts == RTS_VRS:
                self.__model__.vrs = Constraint(rule=self.__vrs_rule(), doc='various return to scale rule')
            
            self.__modeldict[i] = self.__model__

        # Optimize model
    def __NDDF__objective_rule(self):
        """Return the proper objective function"""
        if type(self.b) != type(None):
            def NDDF__objective_rule(model):
                return sum(self.iweight[k]*0 if self.gx[k]==0  else self.iweight[k]*model.thetax[k] for k in model.K
                    ) + sum(self.oweight[l]*0 if self.gy[l]==0 else self.oweight[l]*model.thetay[l] for l in model.L
                    ) + sum(self.bweight[b]*0 if self.gb[b]==0 else self.bweight[b]*model.thetab[b] for b in model.B)
        else:
            def NDDF__objective_rule(model):

                return sum(self.iweight[k]*0 if self.gx[k]==0  else self.iweight[k]*model.thetax[k] for k in model.K
                    ) + sum(self.oweight[l]*0 if self.gy[l]==0 else self.oweight[l]*model.thetay[l] for l in model.L)        
        return NDDF__objective_rule

    def __input_rule(self):
        """Return the proper input constraint"""
        def input_rule(model, k):
            return sum(model.lamda[i2] * self.xref[i2][k] for i2 in model.I2) <= (
                1+ 0*self.gx[k] if self.gx[k]==0 else model.thetax[k]*self.gx[k])*self.x[self.I0][k]
        return input_rule

    def __output_rule(self):
        """Return the proper output constraint"""
        def output_rule(model, l):
            return sum(model.lamda[i2] * self.yref[i2,l] for i2 in model.I2) >= (
                1+0*self.gy[l] if self.gy[l]==0 else model.thetay[l]*self.gy[l])*self.y[self.I0][l]
        return output_rule

    def __undesirable_output_rule(self):
        """Return the proper undesirable output constraint"""
        def undesirable_output_rule(model, b):
            return sum(model.lamda[i2] * self.bref[i2,b] for i2 in model.I2) == (
                1+ 0*self.gb[b]if self.gb[b]==0 else model.thetab[b]*self.gb[b])*self.b[self.I0][b]
        return undesirable_output_rule


    def optimize(self,  solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        
        obj,problem_status,optimization_status,thetax,thetay,thetab,lamda = {},{},{},{},{},{},{}
        for ind, problem in list(self.__modeldict.items()):
            problem_status[ind], optimization_status[ind] = tools.optimize_model(problem, ind, solver)
            
            if type(self.b) != type(None):
                obj[ind],thetax[ind],thetay[ind],thetab[ind],lamda[ind]=problem.objective(),np.asarray(list(problem.thetax[:].value)),np.asarray(list(problem.thetay[:].value)),np.asarray(list(problem.thetab[:].value)),np.asarray(list(problem.lamda[:].value))
            else:
                obj[ind],thetax[ind],thetay[ind],lamda[ind]=problem.objective(),np.asarray(list(problem.thetax[:].value)),np.asarray(list(problem.thetay[:].value)),np.asarray(list(problem.lamda[:].value))

                # print(problem.objective() )

        obj = pd.DataFrame(obj,index=["obj"]).T
        thetax = pd.DataFrame(thetax).T
        thetax.columns = thetax.columns.map(lambda x : "Input"+ str(x)+"'s slack" ) 
        thetay = pd.DataFrame(thetay).T
        thetay.columns = thetay.columns.map(lambda y : "Output"+ str(y)+"'s slack" ) 
        theta = pd.concat([thetax,thetay],axis=1)

        if type(self.b) != type(None):
            thetab = pd.DataFrame(thetab).T
            thetab.columns = thetab.columns.map(lambda b : "Undesirable Output"+ str(b)+"'s slack" ) 
            theta = pd.concat([theta,thetab],axis=1)
        dropcol=[]
        for colnum,g in enumerate(self.gx+self.gy+self.gb if type(self.b) != type(None) else self.gx+self.gy):
            print(g)
            if g==0:
                dropcol.append(theta.columns[colnum])
            
        theta.drop(columns = dropcol,inplace=True)
        
        lamda =pd.DataFrame(lamda) 
        lamda.columns = lamda.columns.map(lambda x : "Dum"+ str(x)+"'s lambda" ) 

        return problem_status,optimization_status,obj,theta,lamda



    def info(self, dmu = "all"):
        """Show the infomation of the lp model

        Args:
            dmu (string): The solver chosen for optimization. Default is "all".
        """
        if dmu =="all":
            for ind, problem in list(self.__modeldict.items()):
                print(ind,"\n",problem.pprint())
 
        print(self.__modeldict[int(dmu)].pprint())





class SBM(DEA):
    def __init__(self,  y, x, b=None, rts=RTS_VRS, yref=None, xref=None, bref=None):
        """DEA: Directional distance function

        Args:
            y (float): output variable. 
            x (float): input variables.
            b (float), optional): undesirable output variables. Defaults to None.
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
            bref (String, optional): reference undesirable output. Defaults to None.
        """
        # Initialize SBM model
        if type(b) != type(None):
            self.y, self.x, self.b,  = tools.assert_valid_mupltiple_y_data_with_bad_outputs(y, x, b,)
        else:
            self.y, self.x,   = tools.assert_valid_mupltiple_y_data(y, x,  )
            self.b=None
        self.rts = rts
        # print(self.b[0])
        if type(yref) != type(None):
            self.yref, self.xref, self.bref = tools.assert_valid_reference_data_with_bad_outputs(self.y, self.x, self.b, yref, xref, bref)
        else:
            self.yref, self.xref, self.bref  = self.y, self.x,self.b 

        self.I = range(len(self.x))          ## I 是 被评价决策单元的数量
        self.__modeldict = {}
        for i in self.I:
            self.I0 = i                                                 ## I 是 被评价决策单元的数量 

            self.__model__ = ConcreteModel()

            # Initialize sets
            self.__model__.I2 = Set(initialize=range(len(self.xref)))      ## I2 是 参考决策单元的数量
            self.__model__.K = Set(initialize=range(len(self.x[0])))          ## K 是投入个数
            self.__model__.L = Set(initialize=range(len(self.y[0])))           ## L 是产出个数 被评价单元和参考单元的K，L一样
            if type(b) != type(None):
                self.__model__.B = Set(initialize=range(len(self.b[0])))   ## B 是 非期望产出个数  
    
            # Initialize variable
            self.__model__.t = Var(Set(initialize=range(1)),domain =PositiveReals,bounds=(0, None), doc='CC object')

            self.__model__.thetax = Var(self.__model__.K,bounds=(0.0, None), doc='slack x')
            self.__model__.thetay = Var(self.__model__.L,bounds=(0.0, None), doc='slack y')
            if type(b) != type(None):
                self.__model__.thetab = Var(self.__model__.B,bounds=(0.0, None), doc='slack b')

            self.__model__.lamda = Var(self.__model__.I2, bounds=(0.0, None), doc='intensity variables')

            # Setup the objective function and constraints
            self.__model__.objective = Objective(rule=self.__objective_rule(), sense=minimize, doc='objective function')
            # self.__model__.cctrans1 = Constraint( expr= self.__model__.t[0] >0, doc='cctrans1')

            self.__model__.cctrans = Constraint(rule=self.__cctrans_rule(), doc='cctrans')
            self.__model__.input = Constraint(self.__model__.K,  rule=self.__input_rule(), doc='input constraint')
            self.__model__.output = Constraint(self.__model__.L,  rule=self.__output_rule(), doc='output constraint')

            if type(b) != type(None):
                self.__model__.undesirable_output = Constraint(self.__model__.B, rule=self.__undesirable_output_rule(), doc='undesirable output constraint')

            if self.rts == RTS_VRS:
                self.__model__.vrs = Constraint(rule=self.__vrs_rule(), doc='various return to scale rule')
            
            self.__modeldict[i] = self.__model__

        # Optimize model
    def __objective_rule(self):
        """Return the proper objective function"""
        def objective_rule(model):
            return model.t[0]-sum(model.thetax[k]/self.x[self.I0,k] for k in model.K) / len(model.K)
        return objective_rule

    def __cctrans_rule(self):
        """Return the cctrans  constraint"""
        if type(self.b) != type(None):
            def cctrans_rule(model):
                return 1==model.t[0]+ (sum(model.thetay[l]/self.y[self.I0,l] for l in model.L) +sum(model.thetay[b]/self.y[self.I0,b] for b in model.B))/(len(model.L)+len(model.B))
        else:
            def cctrans_rule(model):
                return 1==model.t[0]+ (sum(model.thetay[l]/self.y[self.I0,l] for l in model.L) )/(len(model.L))
        return cctrans_rule

    def __input_rule(self):
        """Return the proper input constraint"""
        def input_rule(model, k):
            # print(self.x[self.I0][0]-model.thetax[0])
            return sum(model.lamda[i2] * self.xref[i2][k] for i2 in model.I2) == self.x[self.I0][k]*model.t[0]-model.thetax[k]
        return input_rule

    def __output_rule(self):
        """Return the proper output constraint"""
        def output_rule(model, l):
            return sum(model.lamda[i2] * self.yref[i2][l] for i2 in model.I2) == self.y[self.I0][l]*model.t[0]+model.thetay[l]
        return output_rule

    def __undesirable_output_rule(self):
        """Return the proper undesirable output constraint"""
        def undesirable_output_rule(model, b):
            return sum(model.lamda[i2] * self.bref[i2][b] for i2 in model.I2) == self.b[self.I0][b]*model.t[0]-model.thetab[b]
        return undesirable_output_rule



    def optimize(self,  solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        
        obj,problem_status,optimization_status,thetax,thetay,thetab,lamda = {},{},{},{},{},{},{}
        for ind, problem in list(self.__modeldict.items()):
            problem_status[ind], optimization_status[ind] = tools.optimize_model(problem, ind, solver)
            
            if type(self.b) != type(None):
                obj[ind],thetax[ind],thetay[ind],thetab[ind],lamda[ind]=problem.objective(),np.asarray(list(problem.thetax[:].value))/np.asarray(list(problem.t[:].value)),np.asarray(list(problem.thetay[:].value))/np.asarray(list(problem.t[:].value)),np.asarray(list(problem.thetab[:].value))/np.asarray(list(problem.t[:].value)),np.asarray(list(problem.lamda[:].value))/np.asarray(list(problem.t[:].value))
            else:
                obj[ind],thetax[ind],thetay[ind],lamda[ind]=problem.objective(),np.asarray(list(problem.thetax[:].value))/np.asarray(list(problem.t[:].value)),np.asarray(list(problem.thetay[:].value))/np.asarray(list(problem.t[:].value)),np.asarray(list(problem.lamda[:].value))/np.asarray(list(problem.t[:].value))
                # print(list(problem.thetax[:].value ),list(problem.t[:].value ))
        obj = pd.DataFrame(obj,index=["obj"]).T
        thetax = pd.DataFrame(thetax).T
        thetax.columns = thetax.columns.map(lambda x : "Input"+ str(x)+"'s slack" ) 
        thetay = pd.DataFrame(thetay).T
        thetay.columns = thetay.columns.map(lambda y : "Output"+ str(y)+"'s slack" ) 
        theta=pd.concat([thetax,thetay],axis=1)
        if type(self.b) != type(None):
            thetab = pd.DataFrame(thetab).T
            thetab.columns = thetab.columns.map(lambda b : "Undesirable Output"+ str(b)+"'s slack" ) 
            theta = pd.concat([theta,thetab],axis=1)

        lamda =pd.DataFrame(lamda) 
        lamda.columns = lamda.columns.map(lambda x : "Dum"+ str(x)+"'s lambda" ) 

        return problem_status,optimization_status,obj,theta,lamda
 
    def info(self, dmu = "all"):
        """Show the infomation of the lp model

        Args:
            dmu (string): The solver chosen for optimization. Default is "all".
        """
        if dmu =="all":
            for ind, problem in list(self.__modeldict.items()):
                print(ind,"\n",problem.pprint())
 
        print(self.__modeldict[int(dmu)].pprint())


